from deluca.lung.experimental.controllers._ibonus import IBonus, IBonusCubic, LookaheadPID
# from deluca.lung.experimental.controllers._deep_pid_residual import DeepPIDResidual
# from deluca.lung.experimental.controllers._deep_pid_residual_clipped import DeepPIDResidualClipped
from deluca.lung.experimental.controllers._pid_correction import PIDCorrection
from deluca.lung.experimental.controllers._deep import Deep, deep_train